import abc
from typing import NamedTuple, Optional, cast

import torch
from torch.utils.data import Dataset as TorchDataset

from replay.data.nn import (
    MutableTensorMap,
    SequentialDataset,
    TensorMap,
    TensorSchema,
    TorchSequentialDataset,
    TorchSequentialValidationDataset,
)


class Bert4RecTrainingBatch(NamedTuple):
    """
    Batch of data for training.
    Generated by `Bert4RecTrainingDataset`.
    """

    query_id: torch.LongTensor
    padding_mask: torch.BoolTensor
    features: TensorMap
    tokens_mask: torch.BoolTensor
    labels: torch.LongTensor


class Bert4RecMasker(abc.ABC):
    """
    Interface for a token masking strategy during BERT model training
    """

    @abc.abstractmethod
    def mask(self, paddings: torch.BoolTensor) -> torch.BoolTensor:  # pragma: no cover
        """
        Mask random tokens for only not padded tokens.

        :param paddings: Padding mask where ``0`` is <PAD> and ``1`` otherwise.

        :returns: Mask of sequence where ``0`` is masked and ``1`` otherwise.
        """


class Bert4RecUniformMasker(Bert4RecMasker):
    """
    Token masking strategy that mask random token with uniform distribution.
    """

    def __init__(self, mask_prob: float = 0.15, generator: Optional[torch.Generator] = None) -> None:
        """
        :param mask_prob: Probability of masking each token in sequence.
            Default: ``0.15``.
        :param generator: A pseudorandom number generator for sampling.
            Default: ``None``.
        """
        super().__init__()
        self.generator = generator
        self.mask_prob = mask_prob

    def mask(self, paddings: torch.BoolTensor) -> torch.BoolTensor:
        """
        Mask random token with uniform distribution for only not padded tokens.

        :param paddings: Padding mask where ``0`` is <PAD> and ``1`` otherwise.

        :returns: Mask of sequence where ``0`` is masked and ``1`` otherwise.
        """
        mask_prob = torch.rand(paddings.size(-1), dtype=torch.float32, generator=self.generator)

        # mask[i], 0 ~ mask_prob, 1 ~ (1 - mask_prob)
        mask = (mask_prob * paddings) >= self.mask_prob

        # Fix corner cases in mask
        # 1. If all token are not masked, add mask to the end
        if mask.all():
            mask[-1] = 0
        # 2. If all token are masked, add non-masked before the last
        elif (not mask.any()) and (len(mask) > 1):
            mask[-2] = 1

        return cast(torch.BoolTensor, mask)


class Bert4RecTrainingDataset(TorchDataset):
    """
    Dataset that generates samples to train BERT-like model
    """

    def __init__(
        self,
        sequential: SequentialDataset,
        max_sequence_length: int,
        mask_prob: float = 0.15,
        sliding_window_step: Optional[int] = None,
        label_feature_name: Optional[str] = None,
        custom_masker: Optional[Bert4RecMasker] = None,
        padding_value: Optional[int] = None,
    ) -> None:
        """
        :param sequential: Sequential dataset with training data.
        :param max_sequence_length: Max length of sequence.
        :param mask_prob: Probability of masking each token in sequence.
            Default: ``0.15``.
        :param sliding_window_step: A sliding window step.
            If not ``None`` provides iteration over sequences with window.
            Default: ``None``.
        :param label_feature_name: Name of label feature in provided dataset.
            If ``None`` set an item_id_feature name from sequential dataset.
            Default: ``None``.
        :param custom_masker: Masker object to generate masks for Bert4Rec training.
            If ``None`` set a Bert4RecUniformMasker with provided `mask_prob`.
            Default: ``None``.
        :param padding_value: Value for padding a sequence to match the `max_sequence_length`.
            Default: ``0``.
        """
        super().__init__()
        if label_feature_name:
            if label_feature_name not in sequential.schema:
                msg = "Label feature name not found in provided schema"
                raise ValueError(msg)

            if not sequential.schema[label_feature_name].is_cat:
                msg = "Label feature must be categorical"
                raise ValueError(msg)

            if not sequential.schema[label_feature_name].is_seq:
                msg = "Label feature must be sequential"
                raise ValueError(msg)

        self._max_sequence_length = max_sequence_length
        self._label_feature_name = label_feature_name or sequential.schema.item_id_feature_name
        self._masker = custom_masker or Bert4RecUniformMasker(mask_prob)

        self._inner = TorchSequentialDataset(
            sequential=sequential,
            max_sequence_length=self._max_sequence_length,
            sliding_window_step=sliding_window_step,
            padding_value=padding_value,
        )

    def __len__(self) -> int:
        return len(self._inner)

    def __getitem__(self, index: int) -> Bert4RecTrainingBatch:
        query_id, padding_mask, features = self._inner[index]
        tokens_mask = self._masker.mask(padding_mask)

        assert self._label_feature_name
        labels = features[self._label_feature_name]

        return Bert4RecTrainingBatch(
            query_id=query_id,
            padding_mask=padding_mask,
            features=features,
            tokens_mask=tokens_mask,
            labels=cast(torch.LongTensor, labels),
        )


class Bert4RecPredictionBatch(NamedTuple):
    """
    Batch of data for model inference.
    Generated by `Bert4RecPredictionDataset`.
    """

    query_id: torch.LongTensor
    padding_mask: torch.BoolTensor
    features: TensorMap
    tokens_mask: torch.BoolTensor


class Bert4RecPredictionDataset(TorchDataset):
    """
    Dataset that generates samples to infer BERT-like model
    """

    def __init__(
        self,
        sequential: SequentialDataset,
        max_sequence_length: int,
        padding_value: Optional[int] = None,
    ) -> None:
        """
        :param sequential: Sequential dataset with data to make predictions at.
        :param max_sequence_length: Max length of sequence.
        :param padding_value: Value for padding a sequence to match the `max_sequence_length`.
            Default: ``0``.
        """
        self._schema = sequential.schema
        self._inner = TorchSequentialDataset(
            sequential=sequential,
            max_sequence_length=max_sequence_length,
            padding_value=padding_value,
        )

    def __len__(self) -> int:
        return len(self._inner)

    def __getitem__(self, index: int) -> Bert4RecPredictionBatch:
        query_id, padding_mask, features = self._inner[index]

        shifted_features, shifted_padding_mask, tokens_mask = _shift_features(self._schema, features, padding_mask)

        return Bert4RecPredictionBatch(
            query_id=query_id,
            padding_mask=shifted_padding_mask,
            features=shifted_features,
            tokens_mask=tokens_mask,
        )


class Bert4RecValidationBatch(NamedTuple):
    """
    Batch of data for validation.
    Generated by `Bert4RecValidationDataset`.
    """

    query_id: torch.LongTensor
    padding_mask: torch.BoolTensor
    features: TensorMap
    tokens_mask: torch.BoolTensor
    ground_truth: torch.LongTensor
    train: torch.LongTensor


class Bert4RecValidationDataset(TorchDataset):
    """
    Dataset that generates samples to infer and validate BERT-like model
    """

    def __init__(
        self,
        sequential: SequentialDataset,
        ground_truth: SequentialDataset,
        train: SequentialDataset,
        max_sequence_length: int,
        padding_value: Optional[int] = None,
        label_feature_name: Optional[str] = None,
    ):
        """
        :param sequential: Sequential dataset with data to make predictions at.
        :param ground_truth: Sequential dataset with ground truth predictions.
        :param train: Sequential dataset with training data.
        :param max_sequence_length: Max length of sequence.
        :param padding_value: Value for padding a sequence to match the `max_sequence_length`.
            Default: ``0``.
        :param label_feature_name: Name of label feature in provided dataset.
            If ``None`` set an item_id_feature name from sequential dataset.
            Default: ``None``.
        """
        self._schema = sequential.schema
        self._inner = TorchSequentialValidationDataset(
            sequential=sequential,
            ground_truth=ground_truth,
            max_sequence_length=max_sequence_length,
            padding_value=padding_value,
            label_feature_name=label_feature_name,
            train=train,
        )

    def __len__(self) -> int:
        return len(self._inner)

    def __getitem__(self, index: int) -> Bert4RecValidationBatch:
        query_id, padding_mask, features, ground_truth, train = self._inner[index]

        shifted_features, shifted_padding_mask, tokens_mask = _shift_features(self._schema, features, padding_mask)

        return Bert4RecValidationBatch(
            query_id=query_id,
            padding_mask=shifted_padding_mask,
            features=shifted_features,
            tokens_mask=tokens_mask,
            ground_truth=ground_truth,
            train=train,
        )


def _shift_features(
    schema: TensorSchema,
    features: TensorMap,
    padding_mask: torch.BoolTensor,
) -> tuple[TensorMap, torch.BoolTensor, torch.BoolTensor]:
    shifted_features: MutableTensorMap = {}
    for feature_name, feature in schema.items():
        if feature.is_seq:
            shifted_features[feature_name] = _shift_seq(features[feature_name], feature.padding_value)
        else:
            shifted_features[feature_name] = features[feature_name]

    # [0, 0, 1, 1, 1] -> [0, 1, 1, 1, 0]
    tokens_mask = _shift_seq(padding_mask, 0)

    # [0, 1, 1, 1, 0] -> [0, 1, 1, 1, 1]
    shifted_padding_mask = tokens_mask.clone()
    shifted_padding_mask[-1] = 1

    return (
        shifted_features,
        cast(torch.BoolTensor, shifted_padding_mask),
        cast(torch.BoolTensor, tokens_mask),
    )


def _shift_seq(seq: torch.Tensor, padding_value: int) -> torch.Tensor:
    shifted_seq = seq.roll(-1, dims=0)
    shifted_seq[-1, ...] = padding_value
    return shifted_seq



from collections.abc import Iterable
from importlib.util import find_spec
from typing import Union

from pandas import DataFrame as PandasDataFrame
from polars import DataFrame as PolarsDataFrame
from typing_extensions import TypeAlias


class MissingImport:
    """
    Replacement class with missing import
    """


class FeatureUnavailableError(Exception):
    """Exception class for failing a conditional import check."""


class FeatureUnavailableWarning(Warning):
    """Warning class for failing a conditional import check."""


PYSPARK_AVAILABLE = find_spec("pyspark")
if not PYSPARK_AVAILABLE:
    SparkDataFrame: TypeAlias = MissingImport
else:
    from pyspark.sql import DataFrame

    SparkDataFrame: TypeAlias = DataFrame


TORCH_AVAILABLE = find_spec("torch") and find_spec("lightning")

DataFrameLike = Union[PandasDataFrame, SparkDataFrame, PolarsDataFrame]
IntOrList = Union[Iterable[int], int]
NumType = Union[int, float]


# Conditional import flags
ANN_AVAILABLE = all(
    [
        find_spec("nmslib"),
        find_spec("hnswlib"),
        find_spec("pyarrow"),
    ]
)
OPENVINO_AVAILABLE = TORCH_AVAILABLE and find_spec("onnx") and find_spec("openvino")
OPTUNA_AVAILABLE = find_spec("optuna")
LIGHTFM_AVAILABLE = find_spec("lightfm")



import numpy as np

from .spark_utils import convert2spark
from .types import PYSPARK_AVAILABLE, DataFrameLike

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf


def get_item_recency(
    log: DataFrameLike,
    decay: float = 30,
    limit: float = 0.1,
    kind: str = "exp",
):
    """
    Calculate item weight showing when the majority of interactions with this item happened.

    :param log: interactions log
    :param decay: number of days after which the weight is reduced by half, must be grater than 1
    :param limit: minimal value the weight can reach
    :param kind: type of smoothing, one of [power, exp, linear]
        Corresponding functions are ``power``: ``age^c``,
        ``exp``: ``c^age``, ``linear``: ``1-c*age``
    :return: DataFrame with item weights

    >>> import pandas as pd
    >>> from pyspark.sql.functions import round
    >>> d = {
    ...     "item_idx": [1, 1, 2, 3, 3],
    ...     "timestamp": ["2099-03-19", "2099-03-20", "2099-03-22", "2099-03-27", "2099-03-25"],
    ...     "relevance": [1, 1, 1, 1, 1],
    ... }
    >>> df = pd.DataFrame(d)
    >>> df
       item_idx   timestamp  relevance
    0         1  2099-03-19          1
    1         1  2099-03-20          1
    2         2  2099-03-22          1
    3         3  2099-03-27          1
    4         3  2099-03-25          1

    Age in days is calculated for every item,
    which is transformed into a weight using some function.
    There are three types of smoothing types available: power, exp and linear.
    Each type calculates a parameter ``c`` based on the ``decay`` argument,
    so that an item with ``age==decay`` has weight 0.5.

    Power smoothing falls quickly in the beginning but decays slowly afterwards as ``age^c``.

    >>> (
    ...     get_item_recency(df, kind="power")
    ...     .select("item_idx", "timestamp", round("relevance", 4).alias("relevance"))
    ...     .orderBy("item_idx")
    ...     .show()
    ... )
    +--------+-------------------+---------+
    |item_idx|          timestamp|relevance|
    +--------+-------------------+---------+
    |       1|2099-03-19 12:00:00|   0.6632|
    |       2|2099-03-22 00:00:00|   0.7204|
    |       3|2099-03-26 00:00:00|      1.0|
    +--------+-------------------+---------+
    <BLANKLINE>

    Exponential smoothing is the other way around. Old objects decay more quickly as ``c^age``.

    >>> (
    ...     get_item_recency(df, kind="exp")
    ...     .select("item_idx", "timestamp", round("relevance", 4).alias("relevance"))
    ...     .orderBy("item_idx")
    ...     .show()
    ... )
    +--------+-------------------+---------+
    |item_idx|          timestamp|relevance|
    +--------+-------------------+---------+
    |       1|2099-03-19 12:00:00|   0.8606|
    |       2|2099-03-22 00:00:00|   0.9117|
    |       3|2099-03-26 00:00:00|      1.0|
    +--------+-------------------+---------+
    <BLANKLINE>

    Last type is a linear smoothing: ``1 - c*age``.

    >>> (
    ...     get_item_recency(df, kind="linear")
    ...     .select("item_idx", "timestamp", round("relevance", 4).alias("relevance"))
    ...     .orderBy("item_idx")
    ...     .show()
    ... )
    +--------+-------------------+---------+
    |item_idx|          timestamp|relevance|
    +--------+-------------------+---------+
    |       1|2099-03-19 12:00:00|   0.8917|
    |       2|2099-03-22 00:00:00|   0.9333|
    |       3|2099-03-26 00:00:00|      1.0|
    +--------+-------------------+---------+
    <BLANKLINE>

    This function **does not** take relevance values of interactions into account.
    Only item age is used.
    """
    log = convert2spark(log)
    items = log.select(
        "item_idx",
        sf.unix_timestamp(sf.to_timestamp("timestamp")).alias("timestamp"),
    )
    items = items.groupBy("item_idx").agg(sf.mean("timestamp").alias("timestamp"))
    items = items.withColumn("relevance", sf.lit(1))
    items = smoothe_time(items, decay, limit, kind)
    return items


def smoothe_time(
    log: DataFrameLike,
    decay: float = 30,
    limit: float = 0.1,
    kind: str = "exp",
):
    """
    Weighs ``relevance`` column with a time-dependent weight.

    :param log: interactions log
    :param decay: number of days after which the weight is reduced by half, must be grater than 1
    :param limit: minimal value the weight can reach
    :param kind: type of smoothing, one of [power, exp, linear].
        Corresponding functions are ``power``: ``age^c``,
        ``exp``: ``c^age``, ``linear``: ``1-c*age``
    :return: modified DataFrame

    >>> import pandas as pd
    >>> from pyspark.sql.functions import round
    >>> d = {
    ...     "item_idx": [1, 1, 2, 3, 3],
    ...     "timestamp": ["2099-03-19", "2099-03-20", "2099-03-22", "2099-03-27", "2099-03-25"],
    ...     "relevance": [1, 1, 1, 1, 1],
    ... }
    >>> df = pd.DataFrame(d)
    >>> df
       item_idx   timestamp  relevance
    0         1  2099-03-19          1
    1         1  2099-03-20          1
    2         2  2099-03-22          1
    3         3  2099-03-27          1
    4         3  2099-03-25          1

    Power smoothing falls quickly in the beginning but decays slowly afterwards as ``age^c``.

    >>> (
    ...     smoothe_time(df, kind="power")
    ...     .select("item_idx", "timestamp", round("relevance", 4).alias("relevance"))
    ...     .orderBy("timestamp")
    ...     .show()
    ... )
    +--------+-------------------+---------+
    |item_idx|          timestamp|relevance|
    +--------+-------------------+---------+
    |       1|2099-03-19 00:00:00|    0.639|
    |       1|2099-03-20 00:00:00|   0.6546|
    |       2|2099-03-22 00:00:00|   0.6941|
    |       3|2099-03-25 00:00:00|   0.7994|
    |       3|2099-03-27 00:00:00|      1.0|
    +--------+-------------------+---------+
    <BLANKLINE>

    Exponential smoothing is the other way around. Old objects decay more quickly as ``c^age``.

    >>> (
    ...     smoothe_time(df, kind="exp")
    ...     .select("item_idx", "timestamp", round("relevance", 4).alias("relevance"))
    ...     .orderBy("timestamp")
    ...     .show()
    ... )
    +--------+-------------------+---------+
    |item_idx|          timestamp|relevance|
    +--------+-------------------+---------+
    |       1|2099-03-19 00:00:00|   0.8312|
    |       1|2099-03-20 00:00:00|   0.8507|
    |       2|2099-03-22 00:00:00|   0.8909|
    |       3|2099-03-25 00:00:00|   0.9548|
    |       3|2099-03-27 00:00:00|      1.0|
    +--------+-------------------+---------+
    <BLANKLINE>

    Last type is a linear smoothing: ``1 - c*age``.

    >>> (
    ...     smoothe_time(df, kind="linear")
    ...     .select("item_idx", "timestamp", round("relevance", 4).alias("relevance"))
    ...     .orderBy("timestamp")
    ...     .show()
    ... )
    +--------+-------------------+---------+
    |item_idx|          timestamp|relevance|
    +--------+-------------------+---------+
    |       1|2099-03-19 00:00:00|   0.8667|
    |       1|2099-03-20 00:00:00|   0.8833|
    |       2|2099-03-22 00:00:00|   0.9167|
    |       3|2099-03-25 00:00:00|   0.9667|
    |       3|2099-03-27 00:00:00|      1.0|
    +--------+-------------------+---------+
    <BLANKLINE>

    These examples use constant relevance 1, so resulting weight equals the time dependent weight.
    But actually this value is an updated relevance.

    >>> d = {
    ...     "item_idx": [1, 2, 3],
    ...     "timestamp": ["2099-03-19", "2099-03-20", "2099-03-22"],
    ...     "relevance": [10, 3, 0.1],
    ... }
    >>> df = pd.DataFrame(d)
    >>> df
       item_idx   timestamp  relevance
    0         1  2099-03-19       10.0
    1         2  2099-03-20        3.0
    2         3  2099-03-22        0.1
    >>> (
    ...     smoothe_time(df)
    ...     .select("item_idx", "timestamp", round("relevance", 4).alias("relevance"))
    ...     .orderBy("timestamp")
    ...     .show()
    ... )
    +--------+-------------------+---------+
    |item_idx|          timestamp|relevance|
    +--------+-------------------+---------+
    |       1|2099-03-19 00:00:00|   9.3303|
    |       2|2099-03-20 00:00:00|   2.8645|
    |       3|2099-03-22 00:00:00|      0.1|
    +--------+-------------------+---------+
    <BLANKLINE>
    """
    log = convert2spark(log)
    log = log.withColumn("timestamp", sf.unix_timestamp(sf.to_timestamp("timestamp")))
    last_date = log.agg({"timestamp": "max"}).collect()[0].asDict()["max(timestamp)"]
    day_in_secs = 86400
    log = log.withColumn("age", (last_date - sf.col("timestamp")) / day_in_secs)
    if kind == "power":
        power = np.log(0.5) / np.log(decay)
        log = log.withColumn("age", sf.pow(sf.col("age") + 1, power))
    elif kind == "exp":
        base = np.exp(np.log(0.5) / decay)
        log = log.withColumn("age", sf.pow(base, "age"))
    elif kind == "linear":
        k = 0.5 / decay
        log = log.withColumn("age", 1 - k * sf.col("age"))
    else:
        msg = f"parameter kind must be one of [power, exp, linear], got {kind}"
        raise ValueError(msg)

    log = log.withColumn("age", sf.when(sf.col("age") < limit, limit).otherwise(sf.col("age")))
    log = log.withColumn("relevance", sf.col("relevance") * sf.col("age")).drop("age")
    log = log.withColumn("timestamp", sf.to_timestamp("timestamp"))
    return log



from typing import Optional, Union

from replay.utils import PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_metric import Metric, MetricsDataFrameLike, MetricsReturnType


class Unexpectedness(Metric):
    """
    Fraction of recommended items that are not present in some baseline\
        recommendations.

    .. math::
        Unexpectedness@K(i) = 1 -
            \\frac {\\parallel R^{i}_{1..\\min(K, \\parallel R^{i} \\parallel)}
            \\cap BR^{i}_{1..\\min(K, \\parallel BR^{i} \\parallel)} \\parallel}
            {K}

    .. math::
        Unexpectedness@K = \\frac {1}{N}\\sum_{i=1}^{N}Unexpectedness@K(i)

    :math:`R_{1..j}^{i}` -- the first :math:`j` recommendations for the :math:`i`-th user.

    :math:`BR_{1..j}^{i}` -- the first :math:`j` base recommendations for the :math:`i`-th user.

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> base_rec
       query_id  item_id  rating
    0        1        3    0.5
    1        1        7    0.5
    2        1        2    0.7
    3        2        5    0.6
    4        2        8    0.6
    5        2        3    0.3
    6        3        4    1.0
    7        3        9    0.5
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> Unexpectedness([2, 4])(recommendations, base_rec)
    {'Unexpectedness@2': 0.16666666666666666, 'Unexpectedness@4': 0.5}
    >>> Unexpectedness([2, 4], mode=PerUser())(recommendations, base_rec)
    {'Unexpectedness-PerUser@2': {1: 0.5, 2: 0.0, 3: 0.0},
     'Unexpectedness-PerUser@4': {1: 0.5, 2: 0.5, 3: 0.5}}
    >>> Unexpectedness([2, 4], mode=Median())(recommendations, base_rec)
    {'Unexpectedness-Median@2': 0.0, 'Unexpectedness-Median@4': 0.5}
    >>> Unexpectedness([2, 4], mode=ConfidenceInterval(alpha=0.95))(recommendations, base_rec)
    {'Unexpectedness-ConfidenceInterval@2': 0.32666066409000905,
     'Unexpectedness-ConfidenceInterval@4': 0.0}
    <BLANKLINE>
    """

    def _get_enriched_recommendations(
        self,
        recommendations: Union[PolarsDataFrame, SparkDataFrame],
        base_recommendations: Union[PolarsDataFrame, SparkDataFrame],
    ) -> Union[PolarsDataFrame, SparkDataFrame]:
        if isinstance(recommendations, SparkDataFrame):
            return self._get_enriched_recommendations_spark(recommendations, base_recommendations)
        else:
            return self._get_enriched_recommendations_polars(recommendations, base_recommendations)

    def _get_enriched_recommendations_spark(
        self, recommendations: SparkDataFrame, base_recommendations: SparkDataFrame
    ) -> SparkDataFrame:
        sorted_by_score_recommendations = self._get_items_list_per_user(recommendations)

        sorted_by_score_base_recommendations = self._get_items_list_per_user(base_recommendations).withColumnRenamed(
            "pred_item_id", "base_pred_item_id"
        )

        enriched_recommendations = sorted_by_score_recommendations.join(
            sorted_by_score_base_recommendations, how="left", on=self.query_column
        )

        return self._rearrange_columns(enriched_recommendations)

    def _get_enriched_recommendations_polars(
        self, recommendations: PolarsDataFrame, base_recommendations: PolarsDataFrame
    ) -> PolarsDataFrame:
        sorted_by_score_recommendations = self._get_items_list_per_user(recommendations)

        sorted_by_score_base_recommendations = self._get_items_list_per_user(base_recommendations).rename(
            {"pred_item_id": "base_pred_item_id"}
        )

        enriched_recommendations = sorted_by_score_recommendations.join(
            sorted_by_score_base_recommendations, how="left", on=self.query_column
        )

        return self._rearrange_columns(enriched_recommendations)

    def __call__(
        self,
        recommendations: MetricsDataFrameLike,
        base_recommendations: MetricsDataFrameLike,
    ) -> MetricsReturnType:
        """
        Compute metric.

        :param recommendations: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict): model predictions.
            If DataFrame then it must contains user, item and score columns.
            If dict then key represents user_ids, value represents list of tuple(item_id, score).
        :param base_recommendations: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            base model predictions.
            If DataFrame then it must contains user, item and score columns.
            If dict then key represents user_ids, value represents list of tuple(item_id, score).

        :return: metric values
        """
        self._check_dataframes_equal_types(recommendations, base_recommendations)
        if isinstance(recommendations, SparkDataFrame):
            self._check_duplicates_spark(recommendations)
            self._check_duplicates_spark(base_recommendations)
            assert isinstance(base_recommendations, SparkDataFrame)
            return self._spark_call(recommendations, base_recommendations)
        if isinstance(recommendations, PolarsDataFrame):
            self._check_duplicates_polars(recommendations)
            self._check_duplicates_polars(base_recommendations)
            assert isinstance(base_recommendations, PolarsDataFrame)
            return self._polars_call(recommendations, base_recommendations)
        recommendations = (
            self._convert_pandas_to_dict_with_score(recommendations)
            if isinstance(recommendations, PandasDataFrame)
            else self._convert_dict_to_dict_with_score(recommendations)
        )
        self._check_duplicates_dict(recommendations)
        assert isinstance(base_recommendations, (dict, PandasDataFrame))
        base_recommendations = (
            self._convert_pandas_to_dict_with_score(base_recommendations)
            if isinstance(base_recommendations, PandasDataFrame)
            else self._convert_dict_to_dict_with_score(base_recommendations)
        )
        self._check_duplicates_dict(base_recommendations)

        return self._dict_call(
            list(recommendations),
            recs=recommendations,
            base_recs=base_recommendations,
        )

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], base_recs: Optional[list], recs: Optional[list]) -> list[float]:
        if not base_recs or not recs:
            return [0.0 for _ in ks]
        return [1.0 - len(set(recs[:k]) & set(base_recs[:k])) / k for k in ks]



import warnings
from os.path import join
from typing import Optional, Union

import numpy as np
import pandas as pd
import scipy.sparse as scs
from tqdm import tqdm

from replay.data.dataset import Dataset
from replay.utils import SparkDataFrame
from replay.utils.spark_utils import (
    convert2spark,
    load_pickled_from_parquet,
    save_picklable_to_parquet,
)

from .base_rec import HybridRecommender


class DisjointArm:
    """
    Object for interactions with a single arm in a disjoint LinUCB framework.

    In disjoint LinUCB features of all arms are disjoint (no common features used).
    """

    def __init__(self, arm_index, d, eps, alpha):
        # Track arm index
        self.arm_index = arm_index
        # Exploration parameter
        self.eps = eps
        self.alpha = alpha
        # Inverse of feature matrix for ridge regression
        self.A = self.alpha * np.identity(d)
        self.A_inv = (1.0 / self.alpha) * np.identity(d)
        # right-hand side of the regression
        self.theta = np.zeros(d, dtype=float)

    def feature_update(self, usr_features, relevances):
        """
        Function to update featurs or each Lin-UCB hand in the current model.

        features:
            usr_features = matrix (np.array of shape (m,d)),
                where m = number of occurences of the current feature in the dataset;
            usr_features[i] = features of i-th user, who rated this particular arm (movie);
            relevances = np.array(d) - rating of i-th user, who rated this particular arm (movie);
        """
        # Update A which is (d * d) matrix.
        self.A += np.dot(usr_features.T, usr_features)
        self.A_inv = np.linalg.inv(self.A)
        # Update the parameter theta by the results linear regression
        self.theta = np.linalg.lstsq(self.A, usr_features.T @ relevances, rcond=1.0)[0]


class HybridArm:
    """
    Object for interactions with a single arm in a hybrid LinUCB framework.

    Hybrid LinUCB combines shared and arm-specific features.
    Preferrable when there are meaningful relationships between different arms e.g. genres, product categories, etc.
    """

    def __init__(self, arm_index, d, k, eps, alpha):
        # Track arm index
        self.arm_index = arm_index
        # Exploration parameter
        self.eps = eps
        self.alpha = alpha
        # Inverse of feature matrix for ridge regression
        self.A = scs.csr_matrix(self.alpha * np.identity(d))
        self.A_inv = scs.csr_matrix((1.0 / self.alpha) * np.identity(d))
        self.B = scs.csr_matrix(np.zeros((d, k)))
        # right-hand side of the regression
        self.b = np.zeros(d, dtype=float)

    def feature_update(self, usr_features, usr_itm_features, relevances) -> tuple[np.ndarray, np.ndarray]:
        """
        Function to update featurs or each Lin-UCB hand in the current model.

        features:
            usr_features = matrix (np.array of shape (m,d)),
                where m = number of occurences of the current feature in the dataset;
            usr_features[i] = features of i-th user, who rated this particular arm (movie);
            relevances = np.array(d) - rating of i-th user, who rated this particular arm (movie);
        """

        self.A += (usr_features.T).dot(usr_features)
        self.A_inv = scs.linalg.inv(self.A)
        self.B += (usr_features.T).dot(usr_itm_features)
        self.b += (usr_features.T).dot(relevances)
        delta_A_0 = np.dot(usr_itm_features.T, usr_itm_features) - self.B.T @ self.A_inv @ self.B  # noqa: N806
        delta_b_0 = (usr_itm_features.T).dot(relevances) - (self.B.T).dot(self.A_inv.dot(self.b))
        return delta_A_0, delta_b_0


class LinUCB(HybridRecommender):
    """
    A recommender algorithm for contextual bandit problems.

    Implicitly proposed by `Li et al <https://arxiv.org/pdf/1003.0146>`_.
    The model assumes a linear relationship between user context, item features and action rewards,
    making it efficient for high-dimensional contexts.

    Note: It's recommended to scale features to a similar range (e.g., using StandardScaler or MinMaxScaler)
    to ensure proper convergence and prevent numerical instability (since relationships to learn are linear).

    >>> import pandas as pd
    >>> from replay.data.dataset import (
    ...     Dataset, FeatureHint, FeatureInfo, FeatureSchema, FeatureSource, FeatureType
    ... )
    >>> interactions = pd.DataFrame({"user_id": [0, 1, 2, 2], "item_id": [0, 1, 0, 1], "rating": [1, 0, 0, 0]})
    >>> user_features = pd.DataFrame(
    ...     {"user_id": [0, 1, 2], "usr_feat_1": [1, 2, 3], "usr_feat_2": [4, 5, 6], "usr_feat_3": [7, 8, 9]}
    ... )
    >>> item_features = pd.DataFrame(
    ...     {
    ...         "item_id": [0, 1, 2, 3, 4, 5],
    ...         "itm_feat_1": [1, 2, 3, 4, 5, 6],
    ...         "itm_feat_2": [7, 8, 9, 10, 11, 12],
    ...         "itm_feat_3": [13, 14, 15, 16, 17, 18]
    ...     }
    ... )
    >>> feature_schema = FeatureSchema(
    ...    [
    ...        FeatureInfo(
    ...            column="user_id",
    ...            feature_type=FeatureType.CATEGORICAL,
    ...            feature_hint=FeatureHint.QUERY_ID,
    ...        ),
    ...        FeatureInfo(
    ...            column="item_id",
    ...            feature_type=FeatureType.CATEGORICAL,
    ...            feature_hint=FeatureHint.ITEM_ID,
    ...        ),
    ...        FeatureInfo(
    ...            column="rating",
    ...            feature_type=FeatureType.NUMERICAL,
    ...            feature_hint=FeatureHint.RATING,
    ...        ),
    ...        *[
    ...            FeatureInfo(
    ...                column=name, feature_type=FeatureType.NUMERICAL, feature_source=FeatureSource.ITEM_FEATURES,
    ...            )
    ...            for name in ["itm_feat_1", "itm_feat_2", "itm_feat_3"]
    ...        ],
    ...        *[
    ...            FeatureInfo(
    ...                column=name, feature_type=FeatureType.NUMERICAL, feature_source=FeatureSource.QUERY_FEATURES
    ...            )
    ...            for name in ["usr_feat_1", "usr_feat_2", "usr_feat_3"]
    ...        ],
    ...    ]
    ... )
    >>> dataset = Dataset(
    ...     feature_schema=feature_schema,
    ...     interactions=interactions,
    ...     item_features=item_features,
    ...     query_features=user_features,
    ...     categorical_encoded=True,
    ... )
    >>> dataset.to_spark()
    >>> model = LinUCB(eps=-10.0, alpha=1.0, is_hybrid=False)
    >>> model.fit(dataset)
    >>> model.predict(dataset, k=2, queries=[0,1,2]).toPandas().sort_values(["user_id","rating","item_id"],
    ... ascending=[True,False,True]).reset_index(drop=True)
        user_id   item_id     rating
    0         0         1   -11.073741
    1         0         2   -81.240384
    2         1         0   -6.555529
    3         1         2   -96.436508
    4         2         2   -112.249722
    5         2         3   -112.249722

    """

    _search_space = {
        "eps": {"type": "uniform", "args": [-10.0, 10.0]},
        "alpha": {"type": "uniform", "args": [0.001, 10.0]},
    }
    _study = None  # field required for proper optuna's optimization
    linucb_arms: list[Union[DisjointArm, HybridArm]]  # initialize only when working within fit method
    rel_matrix: np.array  # matrix with relevance scores from predict method
    _num_items: int  # number of items/arms

    def __init__(
        self,
        eps: float,
        alpha: float = 1.0,
        is_hybrid: bool = False,
    ):
        """
        :param eps: exploration coefficient
        :param alpha: ridge parameter
        :param is_hybrid: flag to choose model type. If True, model is hybrid.
        """
        self.is_hybrid = is_hybrid
        self.eps = eps
        self.alpha = alpha

    @property
    def _init_args(self):
        return {"is_hybrid": self.is_hybrid, "eps": self.eps, "alpha": self.alpha}

    def _verify_features(self, dataset: Dataset):
        if dataset.query_features is None:
            msg = "User features are missing"
            raise ValueError(msg)
        if dataset.item_features is None:
            msg = "Item features are missing"
            raise ValueError(msg)
        if (
            len(dataset.feature_schema.query_features.categorical_features) > 0
            or len(dataset.feature_schema.item_features.categorical_features) > 0
        ):
            msg = "Categorical features are not supported"
            raise ValueError(msg)

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        self._verify_features(dataset)

        if not dataset.is_pandas:
            warn_msg = "Dataset will be converted to pandas during internal calculations in fit"
            warnings.warn(warn_msg)
            dataset.to_pandas()

        feature_schema = dataset.feature_schema
        log = dataset.interactions
        user_features = dataset.query_features
        item_features = dataset.item_features

        self._num_items = item_features.shape[0]
        self._user_dim_size = user_features.shape[1] - 1
        self._item_dim_size = item_features.shape[1] - 1
        self._user_idxs_list = set(user_features[feature_schema.query_id_column].values)

        # now initialize an arm object for each potential arm instance
        if self.is_hybrid:
            hybrid_features_k = self._user_dim_size * self._item_dim_size
            self.A_0 = scs.csr_matrix(np.identity(hybrid_features_k))
            self.b_0 = np.zeros(hybrid_features_k, dtype=float)
            self.linucb_arms = [
                HybridArm(
                    arm_index=i,
                    d=self._user_dim_size,
                    k=hybrid_features_k,
                    eps=self.eps,
                    alpha=self.alpha,
                )
                for i in range(self._num_items)
            ]

            for i in tqdm(range(self._num_items)):
                B = log.loc[  # noqa: N806
                    (log[feature_schema.item_id_column] == i)
                    & (log[feature_schema.query_id_column].isin(self._user_idxs_list))
                ]
                if not B.empty:
                    # if we have at least one user interacting with the hand i
                    idxs_list = B[feature_schema.query_id_column].values
                    rel_list = B[feature_schema.interactions_rating_column].values
                    cur_usrs = scs.csr_matrix(
                        user_features.query(f"{feature_schema.query_id_column} in @idxs_list")
                        .drop(columns=[feature_schema.query_id_column])
                        .to_numpy()
                    )
                    cur_itm = scs.csr_matrix(
                        item_features.iloc[i].drop(labels=[feature_schema.item_id_column]).to_numpy()
                    )
                    usr_itm_features = scs.kron(cur_usrs, cur_itm)
                    delta_A_0, delta_b_0 = self.linucb_arms[i].feature_update(  # noqa: N806
                        cur_usrs, usr_itm_features, rel_list
                    )

                    self.A_0 += delta_A_0
                    self.b_0 += delta_b_0

            self.beta = scs.linalg.spsolve(self.A_0, self.b_0)
            self.A_0_inv = scs.linalg.inv(self.A_0)

            for i in range(self._num_items):
                self.linucb_arms[i].theta = scs.linalg.spsolve(
                    self.linucb_arms[i].A,
                    self.linucb_arms[i].b - self.linucb_arms[i].B @ self.beta,
                )
        else:
            self.linucb_arms = [
                DisjointArm(arm_index=i, d=self._user_dim_size, eps=self.eps, alpha=self.alpha)
                for i in range(self._num_items)
            ]

            for i in range(self._num_items):
                B = log.loc[  # noqa: N806
                    (log[feature_schema.item_id_column] == i)
                    & (log[feature_schema.query_id_column].isin(self._user_idxs_list))
                ]
                if not B.empty:
                    # if we have at least one user interacting with the hand i
                    idxs_list = B[feature_schema.query_id_column].values  # noqa: F841
                    rel_list = B[feature_schema.interactions_rating_column].values
                    cur_usrs = user_features.query(f"{feature_schema.query_id_column} in @idxs_list").drop(
                        columns=[feature_schema.query_id_column]
                    )
                    self.linucb_arms[i].feature_update(cur_usrs.to_numpy(), rel_list)

        warn_msg = "Dataset will be converted to spark after internal calculations in fit"
        warnings.warn(warn_msg)
        dataset.to_spark()

    def _predict(
        self,
        dataset: Dataset,
        k: int,
        users: SparkDataFrame,
        items: SparkDataFrame = None,
        filter_seen_items: bool = True,  # noqa: ARG002
        oversample: int = 20,
    ) -> SparkDataFrame:
        self._verify_features(dataset)

        if not dataset.is_pandas:
            warn_msg = "Dataset will be converted to pandas during internal calculations in predict"
            warnings.warn(warn_msg)
            dataset.to_pandas()

        feature_schema = dataset.feature_schema
        user_features = dataset.query_features
        item_features = dataset.item_features
        big_k = min(oversample * k, item_features.shape[0])
        self._user_idxs_list = set(user_features[feature_schema.query_id_column].values)

        users = users.toPandas()
        users = users[users[feature_schema.query_id_column].isin(self._user_idxs_list)]
        num_user_pred = users.shape[0]
        rel_matrix = np.zeros((num_user_pred, self._num_items), dtype=float)

        if self.is_hybrid:
            items = items.toPandas()
            usr_idxs_list = users[feature_schema.query_id_column].values
            itm_idxs_list = items[feature_schema.item_id_column].values  # noqa: F841

            usrs_feat = scs.csr_matrix(
                user_features.query(f"{feature_schema.query_id_column} in @usr_idxs_list")
                .drop(columns=[feature_schema.query_id_column])
                .to_numpy()
            )
            itm_feat = scs.csr_matrix(
                item_features.query(f"{feature_schema.item_id_column} in @itm_idxs_list")
                .drop(columns=[feature_schema.item_id_column])
                .to_numpy()
            )

            # fill in relevance matrix
            for i in tqdm(range(self._num_items)):
                z = scs.kron(usrs_feat, itm_feat[i])
                rel_matrix[:, i] = usrs_feat.dot(self.linucb_arms[i].theta)
                rel_matrix[:, i] += z.dot(self.beta)

                s = (usrs_feat.dot(self.linucb_arms[i].A_inv).multiply(usrs_feat)).sum(axis=1)
                s += (z.dot(self.A_0_inv).multiply(z)).sum(axis=1)
                M = self.A_0_inv @ self.linucb_arms[i].B.T @ self.linucb_arms[i].A_inv  # noqa: N806
                s -= 2 * (z.dot(M).multiply(usrs_feat)).sum(axis=1)
                s += (usrs_feat.dot(M.T @ self.linucb_arms[i].B.T @ self.linucb_arms[i].A_inv).multiply(usrs_feat)).sum(
                    axis=1
                )

                rel_matrix[:, i] += np.array(self.eps * np.sqrt(s))[:, 0]

            # select top k predictions from each row (unsorted ones)
            topk_indices = np.argpartition(rel_matrix, -big_k, axis=1)[:, -big_k:]
            rows_inds, _ = np.indices((num_user_pred, big_k))
            # result df
            predict_inds = np.repeat(usr_idxs_list, big_k)
            predict_items = topk_indices.ravel()
            predict_rels = rel_matrix[rows_inds, topk_indices].ravel()
            # return everything in a PySpark template
            res_df = pd.DataFrame(
                {
                    feature_schema.query_id_column: predict_inds,
                    feature_schema.item_id_column: predict_items,
                    feature_schema.interactions_rating_column: predict_rels,
                }
            )

        else:
            idxs_list = users[feature_schema.query_id_column].values
            usrs_feat = (
                user_features.query(f"{feature_schema.query_id_column} in @idxs_list")
                .drop(columns=[feature_schema.query_id_column])
                .to_numpy()
            )
            # fill in relevance matrix
            for i in range(self._num_items):
                rel_matrix[:, i] = (
                    self.eps * np.sqrt((usrs_feat.dot(self.linucb_arms[i].A_inv) * usrs_feat).sum(axis=1))
                    + usrs_feat @ self.linucb_arms[i].theta
                )
            # select top k predictions from each row (unsorted ones)
            topk_indices = np.argpartition(rel_matrix, -big_k, axis=1)[:, -big_k:]
            rows_inds, _ = np.indices((num_user_pred, big_k))
            # result df
            predict_inds = np.repeat(idxs_list, big_k)
            predict_items = topk_indices.ravel()
            predict_rels = rel_matrix[rows_inds, topk_indices].ravel()
            # return everything in a PySpark template
            res_df = pd.DataFrame(
                {
                    feature_schema.query_id_column: predict_inds,
                    feature_schema.item_id_column: predict_items,
                    feature_schema.interactions_rating_column: predict_rels,
                }
            )

        warn_msg = "Dataset will be converted to spark after internal calculations in predict"
        warnings.warn(warn_msg)
        dataset.to_spark()
        return convert2spark(res_df)

    def _save_model(self, path: str, additional_params: Optional[dict] = None):
        super()._save_model(path, additional_params)

        save_picklable_to_parquet(self.linucb_arms, join(path, "linucb_arms.dump"))

        if self.is_hybrid:
            linucb_hybrid_shared_params = {
                "A_0": self.A_0,
                "A_0_inv": self.A_0_inv,
                "b_0": self.b_0,
                "beta": self.beta,
            }
            save_picklable_to_parquet(
                linucb_hybrid_shared_params,
                join(path, "linucb_hybrid_shared_params.dump"),
            )

    def _load_model(self, path: str):
        super()._load_model(path)

        loaded_linucb_arms = load_pickled_from_parquet(join(path, "linucb_arms.dump"))
        self.linucb_arms = loaded_linucb_arms
        self._num_items = len(loaded_linucb_arms)

        if self.is_hybrid:
            loaded_linucb_hybrid_shared_params = load_pickled_from_parquet(
                join(path, "linucb_hybrid_shared_params.dump")
            )
            for param, value in loaded_linucb_hybrid_shared_params.items():
                setattr(self, param, value)



import json
import os
import pickle
from os.path import join
from pathlib import Path
from typing import Union

from replay.data.dataset_utils import DatasetLabelEncoder
from replay.models.base_rec import BaseRecommender
from replay.splitters import Splitter

from .session_handler import State
from .types import PYSPARK_AVAILABLE

if PYSPARK_AVAILABLE:
    from pyspark.sql import SparkSession

    from .spark_utils import load_pickled_from_parquet, save_picklable_to_parquet

    def get_fs(spark: SparkSession):
        """
        Gets `org.apache.hadoop.fs.FileSystem` instance from JVM gateway

        :param spark: spark session
        :return:
        """
        fs = spark._jvm.org.apache.hadoop.fs.FileSystem.get(spark._jsc.hadoopConfiguration())
        return fs

    def get_list_of_paths(spark: SparkSession, dir_path: str):
        """
        Returns list of paths to files in the `dir_path`

        :param spark: spark session
        :param dir_path: path to dir in hdfs or local disk
        :return: list of paths to files
        """
        fs = get_fs(spark)
        statuses = fs.listStatus(spark._jvm.org.apache.hadoop.fs.Path(dir_path))
        return [str(f.getPath()) for f in statuses]


def save(model: "BaseRecommender", path: Union[str, Path], overwrite: bool = False):
    """
    Save fitted model to disk as a folder

    :param model: Trained recommender
    :param path: destination where model files will be stored
    :return:
    """
    if isinstance(path, Path):
        path = str(path)

    spark = State().session

    fs = get_fs(spark)
    if not overwrite:
        is_exists = fs.exists(spark._jvm.org.apache.hadoop.fs.Path(path))
        if is_exists:
            msg = f"Path '{path}' already exists. Mode is 'overwrite = False'."
            raise FileExistsError(msg)

    fs.mkdirs(spark._jvm.org.apache.hadoop.fs.Path(path))
    model._save_model(join(path, "model"))

    init_args = model._init_args
    init_args["_model_name"] = str(model)
    sc = spark.sparkContext
    df = spark.read.json(sc.parallelize([json.dumps(init_args)]))
    df.coalesce(1).write.mode("overwrite").option("ignoreNullFields", "false").json(join(path, "init_args.json"))

    dataframes = model._dataframes
    df_path = join(path, "dataframes")
    for name, df in dataframes.items():
        if df is not None:
            df.write.mode("overwrite").parquet(join(df_path, name))

    if hasattr(model, "fit_queries"):
        model.fit_queries.write.mode("overwrite").parquet(join(df_path, "fit_queries"))
    if hasattr(model, "fit_items"):
        model.fit_items.write.mode("overwrite").parquet(join(df_path, "fit_items"))
    if hasattr(model, "study"):
        save_picklable_to_parquet(model.study, join(path, "study"))


def load(path: str, model_type=None) -> "BaseRecommender":
    """
    Load saved model from disk

    :param path: path to model folder
    :return: Restored trained model
    """
    # FIXME: Surely there's a better way to handle this? Not having this method at all perhaps?
    import replay.models as models

    spark = State().session
    args = spark.read.json(join(path, "init_args.json")).first().asDict(recursive=True)
    name = args["_model_name"]
    del args["_model_name"]

    model_class = model_type if model_type is not None else getattr(models, name)

    model = model_class(**args)

    dataframes_paths = get_list_of_paths(spark, join(path, "dataframes"))
    for dataframe_path in dataframes_paths:
        df = spark.read.parquet(dataframe_path)
        attr_name = dataframe_path.split("/")[-1]
        setattr(model, attr_name, df)

    model._load_model(join(path, "model"))
    fs = get_fs(spark)
    model.study = (
        load_pickled_from_parquet(join(path, "study"))
        if fs.exists(spark._jvm.org.apache.hadoop.fs.Path(join(path, "study")))
        else None
    )

    return model


def save_encoder(encoder: DatasetLabelEncoder, path: Union[str, Path]) -> None:
    """
    Save fitted DatasetLabelEncoder to disk as a folder

    :param encoder: Trained encoder
    :param path: destination where encoder files will be stored
    """
    if isinstance(path, Path):
        path = str(path)

    os.makedirs(path, exist_ok=True)
    with open(os.path.join(path, "encoder.pickle"), "wb") as cached_file:
        pickle.dump(encoder, cached_file)


def load_encoder(path: Union[str, Path]) -> DatasetLabelEncoder:
    """
    Load saved encoder from disk

    :param path: path to folder
    :return: restored DatasetLabelEncoder
    """
    if isinstance(path, Path):
        path = str(path)

    with open(os.path.join(path, "encoder.pickle"), "rb") as f:
        encoder = pickle.load(f)

    return encoder


def save_splitter(splitter: Splitter, path: str, overwrite: bool = False):
    """
    Save initialized splitter

    :param splitter: Initialized splitter
    :param path: destination where splitter files will be stored
    """
    init_args = splitter._init_args
    init_args["_splitter_name"] = str(splitter)
    spark = State().session
    sc = spark.sparkContext
    df = spark.read.json(sc.parallelize([json.dumps(init_args)]))
    if overwrite:
        df.coalesce(1).write.mode("overwrite").json(join(path, "init_args.json"))
    else:
        df.coalesce(1).write.json(join(path, "init_args.json"))


def load_splitter(path: str) -> Splitter:
    """
    Load splitter

    :param path: path to folder
    :return: restored Splitter
    """
    import replay.splitters as splitters

    spark = State().session
    args = spark.read.json(join(path, "init_args.json")).first().asDict()
    name = args["_splitter_name"]
    del args["_splitter_name"]
    splitter = getattr(splitters, name)
    return splitter(**args)



from dataclasses import dataclass, field
from typing import ClassVar, Optional

from typing_extensions import Literal

from .base_hnsw_param import BaseHnswParam


@dataclass
class NmslibHnswParam(BaseHnswParam):
    """
    Parameters for nmslib-hnsw methods.

    For example,

    >>> NmslibHnswParam(space='negdotprod_sparse',\
                        m=10,\
                        ef_s=200,\
                        ef_c=200,\
                        post=0,\
        )
    NmslibHnswParam(space='negdotprod_sparse', m=10, ef_c=200, post=0, ef_s=200, items_count=None)

    or

    >>> NmslibHnswParam(space='negdotprod_sparse',\
                        m=10,\
                        ef_s=200,\
                        ef_c=200,\
                        post=0,\
        )
    NmslibHnswParam(space='negdotprod_sparse', m=10, ef_c=200, post=0, ef_s=200, items_count=None)

    The reasonable range of values for `m` parameter is 5-100,
    for `ef_c` and `ef_s` is 100-2000.
    Increasing these values improves the prediction quality
    but increases index_time and inference_time too.

     We recommend using these settings:

    - m=16, ef_c=200 and ef_s=200 for simple datasets like MovieLens.
    - m=50, ef_c=1000 and ef_s=1000 for average quality with an average prediction time.
    - m=75, ef_c=2000 and ef_s=2000 for the highest quality with a long prediction time.

    note: choosing these parameters depends on the dataset
    and quality/time tradeoff.

    note: while reducing parameter values the highest range metrics
    like Metric@1000 suffer first.

    note: even in a case with a long training time,
    profit from ann could be obtained while inference will be used multiple times.

    For more details see https://github.com/nmslib/nmslib/blob/master/manual/methods.md.
    """

    space: Literal[
        "cosinesimil_sparse",
        "cosinesimil_sparse_fast",
        "negdotprod_sparse",
        "negdotprod_sparse_fast",
        "angulardist_sparse",
        "angulardist_sparse_fast",
    ] = "negdotprod_sparse_fast"
    items_count: Optional[int] = field(default=None, init=False)

    method: ClassVar[str] = "hnsw"



from dataclasses import dataclass
from enum import Enum

from replay.utils import PYSPARK_AVAILABLE

if PYSPARK_AVAILABLE:
    from pyspark.sql import SparkSession


class FileSystem(Enum):
    """File system types"""

    HDFS = 1
    LOCAL = 2


def get_default_fs() -> str:
    """Returns hadoop `fs.defaultFS` property value"""
    spark = SparkSession.getActiveSession()
    hadoop_conf = spark._jsc.hadoopConfiguration()
    default_fs = hadoop_conf.get("fs.defaultFS")
    return default_fs


@dataclass(frozen=True)
class FileInfo:
    """File meta-information: filesystem, path and hdfs_uri (optional)"""

    path: str
    filesystem: FileSystem
    hdfs_uri: str = None


def get_filesystem(path: str) -> FileInfo:
    """Analyzes path and hadoop config and return `FileInfo` instance with `filesystem`,
    `hdfs uri` (if filesystem is hdfs) and `cleaned path` (without prefix).

    For example:

    >>> path = 'hdfs://node21.bdcl:9000/tmp/file'
    >>> get_filesystem(path)
    FileInfo(path='/tmp/file', filesystem=<FileSystem.HDFS: 1>, hdfs_uri='hdfs://node21.bdcl:9000')

    >>> path = 'file:///tmp/file'
    >>> get_filesystem(path)
    FileInfo(path='/tmp/file', filesystem=<FileSystem.LOCAL: 2>, hdfs_uri=None)

    >>> spark = SparkSession.builder.master("local[1]").getOrCreate()
    >>> path = 'hdfs:///tmp/file'
    >>> get_filesystem(path)
    Traceback (most recent call last):
     ...
    ValueError: Can't get default hdfs uri for path = 'hdfs:///tmp/file'. \
Specify an explicit path, such as 'hdfs://host:port/dir/file', \
or set 'fs.defaultFS' in hadoop configuration.

    >>> path = '/tmp/file'
    >>> get_filesystem(path)
    FileInfo(path='/tmp/file', filesystem=<FileSystem.LOCAL: 2>, hdfs_uri=None)

    >>> spark = SparkSession.builder.master("local[1]").getOrCreate().newSession()
    >>> spark.sparkContext._jsc.hadoopConfiguration().set('fs.defaultFS', 'hdfs://node21.bdcl:9000')
    >>> path = '/tmp/file'
    >>> get_filesystem(path)
    FileInfo(path='/tmp/file', filesystem=<FileSystem.HDFS: 1>, hdfs_uri='hdfs://node21.bdcl:9000')

    >>> path = 'hdfs:///tmp/file'
    >>> get_filesystem(path)
    FileInfo(path='/tmp/file', filesystem=<FileSystem.HDFS: 1>, hdfs_uri='hdfs://node21.bdcl:9000')

    Return fs.defaultFS value because the current session may be used in another test
    >>> spark.sparkContext._jsc.hadoopConfiguration().set('fs.defaultFS', 'file:///')


    Args:
        path (str): path to file on hdfs or local disk

    Returns: FileInfo instance: `filesystem id`,
    `hdfs uri` (if filesystem is hdfs) and `cleaned path` (without prefix)
    """
    prefix_len = 7  # 'hdfs://' and 'file://' length
    if path.startswith("hdfs://"):
        if path.startswith("hdfs:///"):
            default_fs = get_default_fs()
            if default_fs.startswith("hdfs://"):
                return FileInfo(path[prefix_len:], FileSystem.HDFS, default_fs)
            else:
                msg = (
                    f"Can't get default hdfs uri for path = '{path}'. "
                    "Specify an explicit path, such as 'hdfs://host:port/dir/file', "
                    "or set 'fs.defaultFS' in hadoop configuration."
                )
                raise ValueError(msg)
        else:
            hostname = path[prefix_len:].split("/", 1)[0]
            hdfs_uri = "hdfs://" + hostname
            return FileInfo(path[len(hdfs_uri) :], FileSystem.HDFS, hdfs_uri)
    elif path.startswith("file://"):
        return FileInfo(path[prefix_len:], FileSystem.LOCAL)
    else:
        default_fs = get_default_fs()
        if default_fs.startswith("hdfs://"):
            return FileInfo(path, FileSystem.HDFS, default_fs)
        else:
            return FileInfo(path, FileSystem.LOCAL)



"""
This module contains recommender system algorithms including:

- distributed models built in PySpark
- neural networks build in PyTorch with distributed inference in PySpark
- wrappers for commonly used recommender systems libraries and\
models with non-distributed training and distributed inference in PySpark.
"""

from .als import ALSWrap
from .association_rules import AssociationRulesItemRec
from .base_rec import Recommender
from .cat_pop_rec import CatPopRec
from .cluster import ClusterRec
from .kl_ucb import KLUCB
from .knn import ItemKNN
from .lin_ucb import LinUCB
from .pop_rec import PopRec
from .query_pop_rec import QueryPopRec
from .random_rec import RandomRec
from .slim import SLIM
from .thompson_sampling import ThompsonSampling
from .ucb import UCB
from .wilson import Wilson
from .word2vec import Word2VecRec

__all__ = [
    "KLUCB",
    "SLIM",
    "UCB",
    "ALSWrap",
    "AssociationRulesItemRec",
    "CatPopRec",
    "ClusterRec",
    "ItemKNN",
    "LinUCB",
    "PopRec",
    "QueryPopRec",
    "RandomRec",
    "Recommender",
    "ThompsonSampling",
    "Wilson",
    "Word2VecRec",
]



import contextlib
from abc import ABC, abstractmethod
from typing import Optional, Union

import torch
import torch.nn as nn

from replay.data.nn import TensorFeatureInfo, TensorMap, TensorSchema


class Bert4RecModel(torch.nn.Module):
    """
    BERT model
    """

    def __init__(
        self,
        schema: TensorSchema,
        max_len: int = 100,
        hidden_size: int = 256,
        num_blocks: int = 2,
        num_heads: int = 4,
        num_passes_over_block: int = 1,
        dropout: float = 0.1,
        enable_positional_embedding: bool = True,
        enable_embedding_tying: bool = False,
    ) -> None:
        """
        :param schema: Tensor schema of features.
        :param max_len: Max length of sequence.
            Default: ``100``.
        :param hidden_size: Hidden size of transformer.
            Default: ``256``.
        :param num_blocks: Number of Transformer blocks.
            Default: ``2``.
        :param num_heads: Number of Attention heads.
            Default: ``4``.
        :param num_passes_over_block: Number of times to pass data over each Transformer block.
            Default: ``1``.
        :param dropout: Dropout rate.
            Default: ``0.1``.
        :param enable_positional_embedding: Add positional embedding to the result.
            Default: ``True``.
        :param enable_embedding_tying: Use embedding tying head.
            Default: ``False``.
        """
        super().__init__()

        self.schema = schema
        self.max_len = max_len
        self.hidden_size = hidden_size
        self.num_blocks = num_blocks
        self.num_heads = num_heads
        self.num_passes_over_block = num_passes_over_block
        self.dropout = dropout
        self.enable_positional_embedding = enable_positional_embedding
        self.enable_embedding_tying = enable_embedding_tying
        self.item_count = schema.item_id_features.item().cardinality
        assert self.item_count

        self.item_embedder = BertEmbedding(
            schema,
            max_len=max_len,
            dropout=dropout,
            enable_positional_embedding=enable_positional_embedding,
        )

        self.transformer_blocks = torch.nn.ModuleList(
            [
                TransformerBlock(
                    hidden_size,
                    num_heads,
                    4 * hidden_size,
                    dropout,
                )
                for _ in range(num_blocks)
            ]
        )

        self._head: Union[ClassificationHead, EmbeddingTyingHead]
        if self.enable_embedding_tying:
            self._head = EmbeddingTyingHead(self.item_embedder, self.item_count)
        else:
            self._head = ClassificationHead(hidden_size, self.item_count)

        self._init()

    def forward(self, inputs: TensorMap, pad_mask: torch.BoolTensor, token_mask: torch.BoolTensor) -> torch.Tensor:
        """
        :param inputs: Batch of features.
        :param pad_mask: Padding mask where 0 - <PAD>, 1 otherwise.
        :param token_mask: Token mask where 0 - <MASK> tokens, 1 otherwise.

        :returns: Calculated scores.
        """
        output_embeddings = self.forward_step(inputs, pad_mask, token_mask)
        all_scores = self.get_logits(output_embeddings)

        return all_scores  # [B x L x E]

    def predict(
        self,
        inputs: TensorMap,
        pad_mask: torch.BoolTensor,
        token_mask: torch.BoolTensor,
        candidates_to_score: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        :param inputs: Batch of features.
        :param pad_mask: Padding mask where 0 - <PAD>, 1 otherwise.
        :param token_mask: Token mask where 0 - <MASK> tokens, 1 otherwise.
        :param candidates_to_score: Item ids to calculate scores.
            if `None` predicts for all items

        :returns: Calculated scores among canditates_to_score items.
        """
        # final_emb: [B x E]
        final_emb = self.get_query_embeddings(inputs, pad_mask, token_mask)
        candidate_scores = self.get_logits(final_emb, candidates_to_score)
        return candidate_scores

    def forward_step(self, inputs: TensorMap, pad_mask: torch.BoolTensor, token_mask: torch.BoolTensor) -> torch.Tensor:
        """

        :param inputs (TensorMap): Batch of features.
        :param pad_mask (torch.BoolTensor): Padding mask where 0 - <PAD>, 1 otherwise.
        :param token_mask (torch.BoolTensor): Token mask where 0 - <MASK> tokens, 1 otherwise.

        :returns: Output embeddings.
        """

        # B - batch size
        # L - sequence length (max_len)
        # E - embedding size for tokens fed into transformer

        # (B x L x E)
        x = self.item_embedder(inputs, token_mask)

        # Running over multiple transformer blocks
        for transformer in self.transformer_blocks:
            for _ in range(self.num_passes_over_block):
                x = transformer(x, pad_mask)

        return x

    def get_logits(self, out_embeddings: torch.Tensor, item_ids: Optional[torch.LongTensor] = None) -> torch.Tensor:
        """
        Apply head to output embeddings of `forward_step`.

        :param out_embeddings: Embeddings after `forward step`.
        :param item_ids: Item ids to calculate scores.
            Default: ``None``.

        :returns: Logits for each element in `item_ids`.
        """
        return self._head(out_embeddings, item_ids)

    def get_query_embeddings(self, inputs: TensorMap, pad_mask: torch.BoolTensor, token_mask: torch.BoolTensor):
        """
        :param inputs: Batch of features.
        :param pad_mask: Padding mask where 0 - <PAD>, 1 otherwise.
        :param token_mask: Token mask where 0 - <MASK> tokens, 1 otherwise.

        :returns: Query embeddings.
        """
        return self.forward_step(inputs, pad_mask, token_mask)[:, -1, :]

    def _init(self) -> None:
        for _, param in self.named_parameters():
            with contextlib.suppress(ValueError):
                torch.nn.init.xavier_normal_(param.data)


class BertEmbedding(torch.nn.Module):
    """
    BERT Embedding which is consisted with under features
        1. TokenEmbedding : normal embedding matrix
        2. PositionalEmbedding : adding positional information using sin, cos
        sum of all these features are output of BertEmbedding
    """

    def __init__(
        self,
        schema: TensorSchema,
        max_len: int,
        dropout: float = 0.1,
        enable_positional_embedding: bool = True,
        aggregation_method: str = "sum",
    ) -> None:
        """
        :param schema: Tensor schema of features.
        :param max_len: Max length of sequence.
        :param dropout: Dropout rate.
            Default: ``0.1``.
        :param enable_positional_embedding: Add positional embedding to the result.
            Default: ``True``.
        :param aggregation_method: Aggregation method for result embedding.
            Possible values: `"sum"`.
            Default: ``sum``.
        """
        super().__init__()

        self.schema = schema
        self.max_len = max_len
        self.enable_positional_embedding = enable_positional_embedding
        self.aggregation_method = aggregation_method
        self.cat_embeddings = torch.nn.ModuleDict()

        common_dim = None

        for feature_name, tensor_info in schema.items():
            if not tensor_info.is_seq:
                msg = "Non-sequential features is not yet supported"
                raise NotImplementedError(msg)

            dim = tensor_info.embedding_dim if tensor_info.is_cat else tensor_info.tensor_dim

            if aggregation_method == "sum":
                if common_dim is None:
                    common_dim = dim

                if dim != common_dim:
                    msg = "Dimension of all features must be the same for sum aggregation"
                    raise ValueError(msg)
            else:
                raise NotImplementedError()

            if tensor_info.is_cat:
                self.cat_embeddings[feature_name] = CatFeatureEmbedding(tensor_info)

        assert common_dim

        self.embedding_dim = common_dim
        self.dropout = torch.nn.Dropout(p=dropout)
        self.mask_embedding = torch.nn.Embedding(1, common_dim)

        if self.enable_positional_embedding:
            self.position = PositionalEmbedding(max_len=max_len, d_model=common_dim)

    def forward(self, inputs: TensorMap, token_mask: torch.BoolTensor) -> torch.Tensor:
        """
        :param inputs: Batch of features.
        :param token_mask: Token mask where 0 - <MASK> tokens, 1 otherwise.

        :returns: Embeddings for input features.
        """
        if self.aggregation_method == "sum":
            aggregated_embedding: torch.Tensor = None

            for feature_name in self.schema.categorical_features:
                x = inputs[feature_name]
                embedding = self.cat_embeddings[feature_name](x)

                if aggregated_embedding is None:
                    aggregated_embedding = embedding
                else:
                    aggregated_embedding += embedding

            for feature_name in self.schema.numerical_features:
                aggregated_embedding += inputs[feature_name]

        else:
            raise NotImplementedError()

        batch_size = aggregated_embedding.size(0)
        seq_len = aggregated_embedding.size(1)
        embedding_dim = aggregated_embedding.size(2)

        assert seq_len == self.max_len
        assert embedding_dim == self.embedding_dim

        # (B x L) -> (B x L x E)
        #
        # [[0, 1],        [ [[0, 0, 0],
        #  [1, 0]]   -->     [1, 1, 1]],
        #                   [[1, 1, 1],
        #                    [0, 0, 0]] ]
        expanded_mask = token_mask.unsqueeze(-1).expand(-1, -1, embedding_dim)

        # (1 x E) -> (B x L x E)
        #
        # [[1, 2, 3]] -> [ [[1, 2, 3],
        #                   [1, 2, 3]],
        #                  [[1, 2, 3],
        #                   [1, 2, 3]] ]
        expanded_embedding = self.mask_embedding.weight.expand(batch_size, seq_len, -1)

        # Fill masked token embeddings with embedding of [mask] token
        full_embedding = aggregated_embedding.where(expanded_mask, expanded_embedding)

        x = full_embedding
        if self.enable_positional_embedding:
            x += self.position(full_embedding)

        x = self.dropout(x)

        return x

    @property
    def item_embeddings(self) -> torch.Tensor:
        """
        :returns: Item embeddings.
        """
        return self.cat_embeddings[self.schema.item_id_feature_name].weight

    def get_all_embeddings(self) -> dict[str, torch.Tensor]:
        """
        :returns: copy of all embeddings presented in this layer as a dict.
        """
        embeddings = {
            "item_embedding": self.item_embeddings.data.detach().clone(),
        }
        for feature_name in self.schema:
            if feature_name != self.schema.item_id_feature_name:
                embeddings[feature_name] = self.cat_embeddings[feature_name].weight.data.detach().clone()
        if self.enable_positional_embedding:
            embeddings["positional_embedding"] = self.position.pe.weight.data.detach().clone()

        return embeddings


class CatFeatureEmbedding(torch.nn.Embedding):
    """
    Categorical feature embedding.
    """

    def __init__(self, feature: TensorFeatureInfo) -> None:
        """
        :param feature: Categorical tensor feature.
        """
        assert feature.cardinality
        assert feature.embedding_dim
        super().__init__(feature.cardinality, feature.embedding_dim)


class PositionalEmbedding(torch.nn.Module):
    """
    Positional embedding.
    """

    def __init__(self, max_len: int, d_model: int) -> None:
        """
        :param max_len: Max sequence length.
        :param d_model: Embedding dimension.
        """
        super().__init__()
        self.pe = torch.nn.Embedding(max_len, d_model)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: Input embedding.

        :returns: Positional embedding.
        """
        batch_size = x.size(0)
        return self.pe.weight.unsqueeze(0).repeat(batch_size, 1, 1)


class BaseHead(ABC, torch.nn.Module):
    """
    Base abstract head
    """

    def forward(
        self,
        out_embeddings: torch.Tensor,
        item_ids: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        :param out_embeddings: Embeddings after `forward step`.
        :param item_ids: Item ids to calculate scores.
            Default: ``None``.

        :returns: Calculated logits.
        """
        item_embeddings = self.get_item_embeddings()
        bias = self.get_bias()
        if item_ids is not None:
            item_embeddings = item_embeddings[item_ids]
            bias = bias[item_ids]

        logits = torch.nn.functional.linear(out_embeddings, item_embeddings, bias)
        return logits

    @abstractmethod
    def get_item_embeddings(self) -> torch.Tensor:  # pragma: no cover
        """
        :returns: Item embeddings.
        """

    @abstractmethod
    def get_bias(self) -> torch.Tensor:  # pragma: no cover
        """
        :returns: Bias tensor.
        """


class EmbeddingTyingHead(BaseHead):
    """
    Head that calculate logits for all item_ids given output embeddings.
    """

    def __init__(self, item_embedder: BertEmbedding, n_items: int):
        """
        :param item_embedder: Bert embedding.
        :param n_items: Number of items.
        """
        super().__init__()
        self._item_embedder = item_embedder
        self.out_bias = torch.nn.Parameter(torch.Tensor(n_items))
        self.out_bias.data.normal_(0, 0.01)

    def get_item_embeddings(self) -> torch.Tensor:
        """
        :returns: Item embeddings.
        """
        return self._item_embedder.item_embeddings

    def get_bias(self) -> torch.Tensor:
        """
        :returns: Bias tensor.
        """
        return self.out_bias


class ClassificationHead(BaseHead):
    """
    Classification head with linear output
    """

    def __init__(self, hidden_size: int, n_items: int) -> None:
        """
        :param hidden_size: Hidden size of transformer.
        :param n_items: Number of items.
        """
        super().__init__()
        self.linear = torch.nn.Linear(hidden_size, n_items, bias=True)

    def get_item_embeddings(self) -> torch.Tensor:
        """
        :returns: Item embeddings.
        """
        return self.linear.weight

    def get_bias(self) -> torch.Tensor:
        """
        :returns: Bias tensor.
        """
        return self.linear.bias


class TransformerBlock(torch.nn.Module):
    """
    Bidirectional Encoder = Transformer (self-attention)
    Transformer = MultiHead_Attention + Feed_Forward with sublayer connection
    """

    def __init__(
        self,
        hidden_size: int,
        attn_heads: int,
        feed_forward_hidden: int,
        dropout: float,
    ) -> None:
        """
        :param hidden_size: Hidden size of transformer.
        :param attn_heads: Head sizes of multi-head attention.
        :param feed_forward_hidden: Feed_forward_hidden, usually 4*hidden_size.
        :param dropout: Dropout rate.
        """
        super().__init__()
        self.attention = torch.nn.MultiheadAttention(hidden_size, attn_heads, dropout=dropout, batch_first=True)
        self.attention_dropout = torch.nn.Dropout(dropout)
        self.attention_norm = torch.nn.LayerNorm(hidden_size)

        self.pff = PositionwiseFeedForward(d_model=hidden_size, d_ff=feed_forward_hidden, dropout=dropout)
        self.pff_dropout = torch.nn.Dropout(dropout)
        self.pff_norm = torch.nn.LayerNorm(hidden_size)

        self.dropout = torch.nn.Dropout(p=dropout)

    def forward(
        self,
        x: torch.Tensor,
        mask: torch.BoolTensor,
    ) -> torch.Tensor:
        """
        :param x: Input bert embedding.
        :param mask: Mask where 0 - <MASK>, 1 - otherwise.

        :returns: Embedding after Transformer block applied.
        """
        # Attention + skip-connection
        x_norm = self.attention_norm(x)
        attent_emb, _ = self.attention(x_norm, x_norm, x_norm, key_padding_mask=~mask, need_weights=False)
        y = x + self.attention_dropout(attent_emb)

        # PFF + skip-connection
        z = y + self.pff_dropout(self.pff(self.pff_norm(y)))

        return self.dropout(z)


class PositionwiseFeedForward(torch.nn.Module):
    """
    Implements FFN equation.
    """

    def __init__(self, d_model: int, d_ff: int, dropout: float = 0.1) -> None:
        """
        :param d_mode: Embedding dimension.
        :param d_ff: Feed forward dimension, usually 4*d_model.
        :param dropout: Dropout rate.
            Default: ``0.1``.
        """
        super().__init__()
        self.w_1 = torch.nn.Linear(d_model, d_ff)
        self.w_2 = torch.nn.Linear(d_ff, d_model)
        self.dropout = torch.nn.Dropout(dropout)
        self.activation = nn.GELU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: Input tensor.

        :returns: Position wised output.
        """
        return self.w_2(self.dropout(self.activation(self.w_1(x))))



from .base_metric import Metric


class RocAuc(Metric):
    """
    Receiver Operating Characteristic/Area Under the Curve is the aggregated performance measure,
    that depends only on the order of recommended items.
    It can be interpreted as the fraction of object pairs (object of class 1, object of class 0)
    that were correctly ordered by the model.
    The bigger the value of AUC, the better the classification model.

    .. math::
        ROCAUC@K(i) = \\frac {\\sum_{s=1}^{K}\\sum_{t=1}^{K}
        \\mathbb{1}_{r_{si}<r_{ti}}
        \\mathbb{1}_{gt_{si}<gt_{ti}}}
        {\\sum_{s=1}^{K}\\sum_{t=1}^{K} \\mathbb{1}_{gt_{si}<gt_{tj}}}

    :math:`\\mathbb{1}_{r_{si}<r_{ti}}` -- indicator function showing that recommendation score for
    user :math:`i` for item :math:`s` is bigger than for item :math:`t`

    :math:`\\mathbb{1}_{gt_{si}<gt_{ti}}` --  indicator function showing that
    user :math:`i` values item :math:`s` more than item :math:`t`.

    Metric is averaged by all users.

    .. math::
        ROCAUC@K = \\frac {\\sum_{i=1}^{N}ROCAUC@K(i)}{N}

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> groundtruth
       query_id  item_id
    0         1        5
    1         1        6
    2         1        7
    3         1        8
    4         1        9
    5         1       10
    6         2        6
    7         2        7
    8         2        4
    9         2       10
    10        2       11
    11        3        1
    12        3        2
    13        3        3
    14        3        4
    15        3        5
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> RocAuc(2)(recommendations, groundtruth)
    {'RocAuc@2': 0.3333333333333333}
    >>> RocAuc(2, mode=PerUser())(recommendations, groundtruth)
    {'RocAuc-PerUser@2': {1: 0.0, 2: 0.0, 3: 1.0}}
    >>> RocAuc(2, mode=Median())(recommendations, groundtruth)
    {'RocAuc-Median@2': 0.0}
    >>> RocAuc(2, mode=ConfidenceInterval(alpha=0.95))(recommendations, groundtruth)
    {'RocAuc-ConfidenceInterval@2': 0.6533213281800181}
    <BLANKLINE>
    """

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], ground_truth: list, pred: list) -> list[float]:
        if not ground_truth or not pred:
            return [0.0 for _ in ks]
        set_gt = set(ground_truth)
        res = []
        for k in ks:
            length = min(k, len(pred))
            fp_cur = 0
            fp_cum = 0
            for item in pred[:length]:
                if item in set_gt:
                    fp_cum += fp_cur
                else:
                    fp_cur += 1
            if fp_cur == length:
                res.append(0.0)
            elif fp_cum == 0:
                res.append(1.0)
            else:
                res.append(1 - fp_cum / (fp_cur * (length - fp_cur)))
        return res



import abc
from typing import Generic, Optional, Protocol, TypeVar, cast

import lightning
import torch

from replay.models.nn.sequential import Bert4Rec
from replay.models.nn.sequential.postprocessors import BasePostProcessor
from replay.utils import PYSPARK_AVAILABLE, MissingImport, PandasDataFrame, PolarsDataFrame, SparkDataFrame

if PYSPARK_AVAILABLE:  # pragma: no cover
    import pyspark.sql.functions as sf
    from pyspark.sql import SparkSession
    from pyspark.sql.types import ArrayType, DoubleType, IntegerType, StructType
else:
    SparkSession = MissingImport


class PredictionBatch(Protocol):
    """
    Prediction callback batch
    """

    query_id: torch.LongTensor


_T = TypeVar("_T")


class BasePredictionCallback(lightning.Callback, Generic[_T]):
    """
    Base callback for prediction stage
    """

    def __init__(
        self,
        top_k: int,
        query_column: str,
        item_column: str,
        rating_column: str = "rating",
        postprocessors: Optional[list[BasePostProcessor]] = None,
    ) -> None:
        """
        :param top_k: Takes the highest k scores in the ranking.
        :param query_column: query column name.
        :param item_column: item column name.
        :param rating_column: rating column name.
        :param postprocessors: postprocessors to apply.
        """
        super().__init__()
        self.query_column = query_column
        self.item_column = item_column
        self.rating_column = rating_column
        self._top_k = top_k
        self._postprocessors: list[BasePostProcessor] = postprocessors or []
        self._query_batches: list[torch.Tensor] = []
        self._item_batches: list[torch.Tensor] = []
        self._item_scores: list[torch.Tensor] = []

    def on_predict_epoch_start(
        self, trainer: lightning.Trainer, pl_module: lightning.LightningModule  # noqa: ARG002
    ) -> None:
        self._query_batches.clear()
        self._item_batches.clear()
        self._item_scores.clear()

        candidates = trainer.model.candidates_to_score if hasattr(trainer.model, "candidates_to_score") else None
        for postprocessor in self._postprocessors:
            if hasattr(postprocessor, "candidates"):
                postprocessor.candidates = candidates

    def on_predict_batch_end(
        self,
        trainer: lightning.Trainer,  # noqa: ARG002
        pl_module: lightning.LightningModule,  # noqa: ARG002
        outputs: torch.Tensor,
        batch: PredictionBatch,
        batch_idx: int,  # noqa: ARG002
        dataloader_idx: int = 0,  # noqa: ARG002
    ) -> None:
        query_ids, scores = self._compute_pipeline(batch.query_id, outputs)
        top_scores, top_item_ids = torch.topk(scores, k=self._top_k, dim=1)
        self._query_batches.append(query_ids)
        self._item_batches.append(top_item_ids)
        self._item_scores.append(top_scores)

    def get_result(self) -> _T:
        """
        :returns: prediction result
        """
        prediction = self._ids_to_result(
            torch.cat(self._query_batches),
            torch.cat(self._item_batches),
            torch.cat(self._item_scores),
        )
        return prediction

    def _compute_pipeline(
        self, query_ids: torch.LongTensor, scores: torch.Tensor
    ) -> tuple[torch.LongTensor, torch.Tensor]:
        for postprocessor in self._postprocessors:
            query_ids, scores = postprocessor.on_prediction(query_ids, scores)
        return query_ids, scores

    @abc.abstractmethod
    def _ids_to_result(
        self,
        query_ids: torch.Tensor,
        item_ids: torch.Tensor,
        item_scores: torch.Tensor,
    ) -> _T:  # pragma: no cover
        pass


class PandasPredictionCallback(BasePredictionCallback[PandasDataFrame]):
    """
    Callback for predition stage with pandas data frame
    """

    def _ids_to_result(
        self,
        query_ids: torch.Tensor,
        item_ids: torch.Tensor,
        item_scores: torch.Tensor,
    ) -> PandasDataFrame:
        prediction = PandasDataFrame(
            {
                self.query_column: query_ids.flatten().cpu().numpy(),
                self.item_column: list(item_ids.cpu().numpy()),
                self.rating_column: list(item_scores.cpu().numpy()),
            }
        )
        return prediction.explode([self.item_column, self.rating_column])


class PolarsPredictionCallback(BasePredictionCallback[PolarsDataFrame]):
    """
    Callback for predition stage with polars data frame
    """

    def _ids_to_result(
        self,
        query_ids: torch.Tensor,
        item_ids: torch.Tensor,
        item_scores: torch.Tensor,
    ) -> PolarsDataFrame:
        prediction = PolarsDataFrame(
            {
                self.query_column: query_ids.flatten().cpu().numpy(),
                self.item_column: list(item_ids.cpu().numpy()),
                self.rating_column: list(item_scores.cpu().numpy()),
            }
        )
        return prediction.explode([self.item_column, self.rating_column])


class SparkPredictionCallback(BasePredictionCallback[SparkDataFrame]):
    """
    Callback for prediction stage with spark data frame
    """

    def __init__(
        self,
        top_k: int,
        query_column: str,
        item_column: str,
        rating_column: str,
        spark_session: SparkSession,
        postprocessors: Optional[list[BasePostProcessor]] = None,
    ) -> None:
        """
        :param top_k: Takes the highest k scores in the ranking.
        :param query_column: query column name.
        :param item_column: item column name.
        :param rating_column: rating column name.
        :param postprocessors: postprocessors to apply.
        """
        super().__init__(
            top_k=top_k,
            query_column=query_column,
            item_column=item_column,
            rating_column=rating_column,
            postprocessors=postprocessors,
        )
        self.spark_session = spark_session

    def _ids_to_result(
        self,
        query_ids: torch.Tensor,
        item_ids: torch.Tensor,
        item_scores: torch.Tensor,
    ) -> SparkDataFrame:
        schema = (
            StructType()
            .add(self.query_column, IntegerType(), False)
            .add(self.item_column, ArrayType(IntegerType()), False)
            .add(self.rating_column, ArrayType(DoubleType()), False)
        )
        prediction = (
            self.spark_session.createDataFrame(
                data=list(
                    zip(
                        query_ids.flatten().cpu().numpy().tolist(),
                        item_ids.cpu().numpy().tolist(),
                        item_scores.cpu().numpy().tolist(),
                    )
                ),
                schema=schema,
            )
            .withColumn("exploded_columns", sf.explode(sf.arrays_zip(self.item_column, self.rating_column)))
            .select(self.query_column, f"exploded_columns.{self.item_column}", f"exploded_columns.{self.rating_column}")
        )
        return prediction


class TorchPredictionCallback(BasePredictionCallback[tuple[torch.LongTensor, torch.LongTensor, torch.Tensor]]):
    """
    Callback for predition stage with tuple of tensors
    """

    def __init__(
        self,
        top_k: int,
        postprocessors: Optional[list[BasePostProcessor]] = None,
    ) -> None:
        """
        :param top_k: Takes the highest k scores in the ranking.
        :param postprocessors: postprocessors to apply.
        """
        super().__init__(
            top_k=top_k,
            query_column="query_id",
            item_column="item_id",
            rating_column="rating",
            postprocessors=postprocessors,
        )

    def _ids_to_result(
        self,
        query_ids: torch.Tensor,
        item_ids: torch.Tensor,
        item_scores: torch.Tensor,
    ) -> tuple[torch.LongTensor, torch.LongTensor, torch.Tensor]:
        return (
            cast(torch.LongTensor, query_ids.flatten().cpu().long()),
            cast(torch.LongTensor, item_ids.cpu().long()),
            item_scores.cpu(),
        )


class QueryEmbeddingsPredictionCallback(lightning.Callback):
    """
    Callback for prediction stage to get query embeddings.
    """

    def __init__(self):
        self._embeddings_per_batch: list[torch.Tensor] = []

    def on_predict_epoch_start(
        self, trainer: lightning.Trainer, pl_module: lightning.LightningModule  # noqa: ARG002
    ) -> None:
        self._embeddings_per_batch.clear()

    def on_predict_batch_end(
        self,
        trainer: lightning.Trainer,  # noqa: ARG002
        pl_module: lightning.LightningModule,
        outputs: torch.Tensor,  # noqa: ARG002
        batch: PredictionBatch,
        batch_idx: int,  # noqa: ARG002
        dataloader_idx: int = 0,  # noqa: ARG002
    ) -> None:
        args = [batch.features, batch.padding_mask]
        if isinstance(pl_module, Bert4Rec):
            args.append(batch.tokens_mask)

        query_embeddings = pl_module._model.get_query_embeddings(*args)
        self._embeddings_per_batch.append(query_embeddings)

    def get_result(self):
        """
        :returns: Query embeddings through all batches.
        """
        return torch.cat(self._embeddings_per_batch)



from typing import Literal, Optional

import numpy as np
import pandas as pd
import polars as pl

from replay.utils import (
    PYSPARK_AVAILABLE,
    DataFrameLike,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)

from .base_splitter import Splitter

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql import Window

StrategyName = Literal["interactions", "timedelta"]


class LastNSplitter(Splitter):
    """
    Split interactions by last N interactions/timedelta per user.
    Type of splitting depends on the ``strategy`` parameter.

    >>> from datetime import datetime
    >>> import pandas as pd
    >>> columns = ["query_id", "item_id", "timestamp"]
    >>> data = [
    ...     (1, 1, "01-01-2020"),
    ...     (1, 2, "02-01-2020"),
    ...     (1, 3, "03-01-2020"),
    ...     (1, 4, "04-01-2020"),
    ...     (1, 5, "05-01-2020"),
    ...     (2, 1, "06-01-2020"),
    ...     (2, 2, "07-01-2020"),
    ...     (2, 3, "08-01-2020"),
    ...     (2, 9, "09-01-2020"),
    ...     (2, 10, "10-01-2020"),
    ...     (3, 1, "01-01-2020"),
    ...     (3, 5, "02-01-2020"),
    ...     (3, 3, "03-01-2020"),
    ...     (3, 1, "04-01-2020"),
    ...     (3, 2, "05-01-2020"),
    ... ]
    >>> dataset = pd.DataFrame(data, columns=columns)
    >>> dataset["timestamp"] = pd.to_datetime(dataset["timestamp"], format="%d-%m-%Y")
    >>> dataset
       query_id  item_id  timestamp
    0         1        1 2020-01-01
    1         1        2 2020-01-02
    2         1        3 2020-01-03
    3         1        4 2020-01-04
    4         1        5 2020-01-05
    5         2        1 2020-01-06
    6         2        2 2020-01-07
    7         2        3 2020-01-08
    8         2        9 2020-01-09
    9         2       10 2020-01-10
    10        3        1 2020-01-01
    11        3        5 2020-01-02
    12        3        3 2020-01-03
    13        3        1 2020-01-04
    14        3        2 2020-01-05
    >>> splitter = LastNSplitter(
    ...     N=2,
    ...     divide_column="query_id",
    ...     time_column_format="yyyy-MM-dd",
    ...     query_column="query_id",
    ...     item_column="item_id"
    ... )
    >>> train, test = splitter.split(dataset)
    >>> train
       query_id  item_id  timestamp
    0         1        1 2020-01-01
    1         1        2 2020-01-02
    2         1        3 2020-01-03
    5         2        1 2020-01-06
    6         2        2 2020-01-07
    7         2        3 2020-01-08
    10        3        1 2020-01-01
    11        3        5 2020-01-02
    12        3        3 2020-01-03
    >>> test
       query_id  item_id  timestamp
    3         1        4 2020-01-04
    4         1        5 2020-01-05
    8         2        9 2020-01-09
    9         2       10 2020-01-10
    13        3        1 2020-01-04
    14        3        2 2020-01-05
    <BLANKLINE>
    """

    _init_arg_names = [
        "N",
        "divide_column",
        "time_column_format",
        "strategy",
        "drop_cold_users",
        "drop_cold_items",
        "query_column",
        "item_column",
        "timestamp_column",
        "session_id_column",
        "session_id_processing_strategy",
    ]

    def __init__(
        self,
        N: int,  # noqa: N803
        divide_column: str = "query_id",
        time_column_format: str = "yyyy-MM-dd HH:mm:ss",
        strategy: StrategyName = "interactions",
        drop_cold_users: bool = False,
        drop_cold_items: bool = False,
        query_column: str = "query_id",
        item_column: str = "item_id",
        timestamp_column: str = "timestamp",
        session_id_column: Optional[str] = None,
        session_id_processing_strategy: str = "test",
    ):
        """
        :param N: Number of last interactions or size of the time window in seconds
        :param divide_column: Name of column for dividing
            in dataframe, default: ``query_id``.
        :param time_column_format: Format of the timestamp column,
            used for converting string dates to a numerical timestamp when strategy is 'timedelta'.
            If the column is already a datetime object or a numerical timestamp, this parameter is ignored.
            default: ``yyyy-MM-dd HH:mm:ss``
        :param strategy: Defines the type of data splitting.
            Must be ``interactions`` or ``timedelta``.
            default: ``interactions``.
        :param query_column: Name of query interaction column.
        :param drop_cold_users: Drop users from test DataFrame.
            which are not in train DataFrame, default: False.
        :param drop_cold_items: Drop items from test DataFrame
            which are not in train DataFrame, default: False.
        :param item_column: Name of item interaction column.
            If ``drop_cold_items`` is ``False``, then you can omit this parameter.
            Default: ``item_id``.
        :param timestamp_column: Name of time column,
            Default: ``timestamp``.
        :param session_id_column: Name of session id column, which values can not be split,
            default: ``None``.
        :param session_id_processing_strategy: strategy of processing session if it is split,
            values: ``train, test``, train: whole split session goes to train. test: same but to test.
            default: ``test``.
        """
        if strategy not in ["interactions", "timedelta"]:
            msg = "strategy must be equal 'interactions' or 'timedelta'"
            raise ValueError(msg)
        super().__init__(
            drop_cold_users=drop_cold_users,
            drop_cold_items=drop_cold_items,
            query_column=query_column,
            item_column=item_column,
            timestamp_column=timestamp_column,
            session_id_column=session_id_column,
            session_id_processing_strategy=session_id_processing_strategy,
        )
        self.N = N
        self.strategy = strategy
        self.divide_column = divide_column
        self.time_column_format = None
        if self.strategy == "timedelta":
            self.time_column_format = time_column_format

    def _add_time_partition(self, interactions: DataFrameLike) -> DataFrameLike:
        if isinstance(interactions, SparkDataFrame):
            return self._add_time_partition_to_spark(interactions)
        if isinstance(interactions, PandasDataFrame):
            return self._add_time_partition_to_pandas(interactions)
        if isinstance(interactions, PolarsDataFrame):
            return self._add_time_partition_to_polars(interactions)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)

    def _add_time_partition_to_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        res = interactions.copy(deep=True)
        res.sort_values(by=[self.divide_column, self.timestamp_column], inplace=True)
        res["row_num"] = res.groupby(self.divide_column, sort=False).cumcount() + 1

        return res

    def _add_time_partition_to_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        res = interactions.withColumn(
            "row_num",
            sf.row_number().over(Window.partitionBy(self.divide_column).orderBy(sf.col(self.timestamp_column))),
        )

        return res

    def _add_time_partition_to_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        res = interactions.sort(self.timestamp_column).with_columns(
            pl.col(self.divide_column).cum_count().over(pl.col(self.divide_column)).alias("row_num")
        )

        return res

    def _to_unix_timestamp(self, interactions: DataFrameLike) -> DataFrameLike:
        if isinstance(interactions, SparkDataFrame):
            return self._to_unix_timestamp_spark(interactions)
        if isinstance(interactions, PandasDataFrame):
            return self._to_unix_timestamp_pandas(interactions)
        if isinstance(interactions, PolarsDataFrame):
            return self._to_unix_timestamp_polars(interactions)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)

    def _to_unix_timestamp_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        time_column_type = dict(interactions.dtypes)[self.timestamp_column]
        if time_column_type == np.dtype("datetime64[ns]"):
            interactions = interactions.copy(deep=True)
            interactions[self.timestamp_column] = (
                interactions[self.timestamp_column] - pd.Timestamp("1970-01-01")
            ) // pd.Timedelta("1s")

        return interactions

    def _to_unix_timestamp_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        time_column_type = dict(interactions.dtypes)[self.timestamp_column]
        if time_column_type == "date":
            interactions = interactions.withColumn(
                self.timestamp_column,
                sf.unix_timestamp(self.timestamp_column, self.time_column_format),
            )

        return interactions

    def _to_unix_timestamp_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        time_column_type = interactions.dtypes[interactions.get_column_index(self.timestamp_column)]
        if isinstance(time_column_type, pl.Datetime):
            interactions = interactions.with_columns(pl.col(self.timestamp_column).dt.epoch("s"))

        return interactions

    def _partial_split_interactions(self, interactions: DataFrameLike, n: int) -> tuple[DataFrameLike, DataFrameLike]:
        res = self._add_time_partition(interactions)
        if isinstance(interactions, SparkDataFrame):
            return self._partial_split_interactions_spark(res, n)
        if isinstance(interactions, PandasDataFrame):
            return self._partial_split_interactions_pandas(res, n)
        return self._partial_split_interactions_polars(res, n)

    def _partial_split_interactions_pandas(
        self, interactions: PandasDataFrame, n: int
    ) -> tuple[PandasDataFrame, PandasDataFrame]:
        interactions["count"] = interactions.groupby(self.divide_column, sort=False)[self.divide_column].transform(len)
        interactions["is_test"] = interactions["row_num"] > (interactions["count"] - float(n))
        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions[~interactions["is_test"]].drop(columns=["row_num", "count", "is_test"])
        test = interactions[interactions["is_test"]].drop(columns=["row_num", "count", "is_test"])

        return train, test

    def _partial_split_interactions_spark(
        self, interactions: SparkDataFrame, n: int
    ) -> tuple[SparkDataFrame, SparkDataFrame]:
        interactions = interactions.withColumn(
            "count",
            sf.count(self.timestamp_column).over(Window.partitionBy(self.divide_column)),
        )
        # float(n) - because DataFrame.filter is changing order
        # of sorted DataFrame to descending
        interactions = interactions.withColumn("is_test", sf.col("row_num") > sf.col("count") - sf.lit(float(n)))
        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions.filter("is_test == 0").drop("row_num", "count", "is_test")
        test = interactions.filter("is_test").drop("row_num", "count", "is_test")

        return train, test

    def _partial_split_interactions_polars(
        self, interactions: PolarsDataFrame, n: int
    ) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        interactions = interactions.with_columns(
            pl.col(self.timestamp_column).count().over(self.divide_column).alias("count")
        )
        interactions = interactions.with_columns((pl.col("row_num") > (pl.col("count") - n)).alias("is_test"))
        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions.filter(~pl.col("is_test")).drop("row_num", "count", "is_test")
        test = interactions.filter(pl.col("is_test")).drop("row_num", "count", "is_test")

        return train, test

    def _partial_split_timedelta(
        self, interactions: DataFrameLike, timedelta: int
    ) -> tuple[DataFrameLike, DataFrameLike]:
        if isinstance(interactions, SparkDataFrame):
            return self._partial_split_timedelta_spark(interactions, timedelta)
        if isinstance(interactions, PandasDataFrame):
            return self._partial_split_timedelta_pandas(interactions, timedelta)
        return self._partial_split_timedelta_polars(interactions, timedelta)

    def _partial_split_timedelta_pandas(
        self, interactions: PandasDataFrame, timedelta: int
    ) -> tuple[PandasDataFrame, PandasDataFrame]:
        res = interactions.copy(deep=True)
        res["diff_timestamp"] = (
            res.groupby(self.divide_column)[self.timestamp_column].transform(max) - res[self.timestamp_column]
        )
        res["is_test"] = res["diff_timestamp"] < timedelta
        if self.session_id_column:
            res = self._recalculate_with_session_id_column(res)

        train = res[~res["is_test"]].drop(columns=["diff_timestamp", "is_test"])
        test = res[res["is_test"]].drop(columns=["diff_timestamp", "is_test"])

        return train, test

    def _partial_split_timedelta_spark(
        self, interactions: SparkDataFrame, timedelta: int
    ) -> tuple[SparkDataFrame, SparkDataFrame]:
        inter_with_max_time = interactions.withColumn(
            "max_timestamp",
            sf.max(self.timestamp_column).over(Window.partitionBy(self.divide_column)),
        )
        inter_with_diff = inter_with_max_time.withColumn(
            "diff_timestamp", sf.col("max_timestamp") - sf.col(self.timestamp_column)
        )
        # drop unnecessary column
        inter_with_diff = inter_with_diff.drop("max_timestamp")

        res = inter_with_diff.withColumn("is_test", sf.col("diff_timestamp") < sf.lit(timedelta))
        if self.session_id_column:
            res = self._recalculate_with_session_id_column(res)

        train = res.filter("is_test == 0").drop("diff_timestamp", "is_test")
        test = res.filter("is_test").drop("diff_timestamp", "is_test")

        return train, test

    def _partial_split_timedelta_polars(
        self, interactions: PolarsDataFrame, timedelta: int
    ) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        res = interactions.with_columns(
            (pl.col(self.timestamp_column).max().over(self.divide_column) - pl.col(self.timestamp_column)).alias(
                "diff_timestamp"
            )
        ).with_columns((pl.col("diff_timestamp") < timedelta).alias("is_test"))

        if self.session_id_column:
            res = self._recalculate_with_session_id_column(res)

        train = res.filter(~pl.col("is_test")).drop("diff_timestamp", "is_test")
        test = res.filter(pl.col("is_test")).drop("diff_timestamp", "is_test")

        return train, test

    def _core_split(self, interactions: DataFrameLike) -> list[DataFrameLike]:
        if self.strategy == "timedelta":
            interactions = self._to_unix_timestamp(interactions)
        train, test = getattr(self, "_partial_split_" + self.strategy)(interactions, self.N)

        return train, test



from .prediction_callbacks import (
    BasePredictionCallback,
    PandasPredictionCallback,
    PolarsPredictionCallback,
    QueryEmbeddingsPredictionCallback,
    SparkPredictionCallback,
    TorchPredictionCallback,
)
from .validation_callback import ValidationMetricsCallback
